import os

from sklearn.datasets import make_classification
from sklearn.model_selection import ShuffleSplit
from sklearn.metrics.pairwise import rbf_kernel
from joblib import Memory

from photonai import Hyperpipe, PipelineElement, FloatRange


cache_dir = './tmp/kernel_cache'
os.makedirs(cache_dir, exist_ok=True)
memory = Memory(cachedir=cache_dir, verbose=0)


@memory.cache
def cached_rbf(X, Y):
    return rbf_kernel(X, Y)


# create toy data
n_features = 10000
n_samples = 1000
n_informative = 10
X, y = make_classification(n_samples, n_features, n_informative=n_informative)
gamma = 1 / n_features

"""
Especially with large datasets, it is unnecessary to recompute the kernel for every hyperparameter configuration.
For that reason, you can pass a cached kernel function that will only recompute the kernel if the input data changes.
If you don't want to cache the kernel, it still decreases the computation time by magnitudes when passing the kernel
as dedicated function. See this issue for details:  
https://github.com/scikit-learn/scikit-learn/issues/21410
https://stackoverflow.com/questions/69680420/using-a-custom-rbf-kernel-function-for-sklearns-svc-is-way-faster-than-built-in
"""
#kernel = 'kernel'
#kernel = rbf_kernel
kernel = cached_rbf

pipe = Hyperpipe('svm_with_custom_kernel',
                 inner_cv=ShuffleSplit(n_splits=1, test_size=0.2),
                 outer_cv=ShuffleSplit(n_splits=1, test_size=0.2),
                 optimizer='sk_opt',
                 optimizer_params={'n_configurations': 15},
                 metrics=['accuracy', 'precision', 'recall', 'balanced_accuracy'],
                 best_config_metric='accuracy',
                 project_folder='./tmp',
                 verbosity=1)

pipe += PipelineElement('StandardScaler')

pipe += PipelineElement('SVC',
                        hyperparameters={'C': FloatRange(1e-6, 1e6)},
                        gamma=gamma, kernel=kernel)

pipe.fit(X, y)




from sklearn.datasets import load_breast_cancer
from sklearn.model_selection import KFold

from photonai import Hyperpipe, PipelineElement, Preprocessing, FloatRange, Categorical, IntegerRange

# WE USE THE BREAST CANCER SET FROM SKLEARN
X, y = load_breast_cancer(return_X_y=True)

# DESIGN YOUR PIPELINE
my_pipe = Hyperpipe('batching_pipe',
                    optimizer='grid_search',
                    metrics=['accuracy', 'precision', 'recall', 'balanced_accuracy'],
                    best_config_metric='accuracy',
                    project_folder='./tmp/',
                    outer_cv=KFold(n_splits=3),
                    inner_cv=KFold(n_splits=5),
                    verbosity=0)


preprocessing = Preprocessing()
preprocessing += PipelineElement("LabelEncoder", batch_size=10)

my_pipe += preprocessing

# ADD ELEMENTS TO YOUR PIPELINE
# first normalize all features
my_pipe.add(PipelineElement('StandardScaler'))
# then do feature selection using a PCA, specify which values to try in the hyperparameter search
my_pipe += PipelineElement('PCA', hyperparameters={'n_components': IntegerRange(5, 10)}, test_disabled=True)
# engage and optimize the good old SVM for Classification
my_pipe += PipelineElement('SVC', hyperparameters={'kernel': Categorical(['rbf', 'linear']),
                                                   'C': FloatRange(0.5, 1)}, gamma='scale')

# NOW TRAIN YOUR PIPELINE
my_pipe.fit(X, y)

debug = True



from sklearn.datasets import load_diabetes
from photonai import RegressionPipe

my_pipe = RegressionPipe('diabetes',
                         add_default_pipeline_elements=True,
                         scaling=True,
                         imputation=False,
                         imputation_nan_value=None,
                         feature_selection=False,
                         dim_reduction=True,
                         n_pca_components=10,
                         add_estimator=True)
# load data and train
X, y = load_diabetes(return_X_y=True)
my_pipe.fit(X, y)



from photonai import ClassificationPipe
from sklearn.datasets import load_breast_cancer
from sklearn.model_selection import ShuffleSplit


X, y = load_breast_cancer(return_X_y=True)
my_pipe = ClassificationPipe(name='breast_cancer_analysis',
                             inner_cv=ShuffleSplit(n_splits=2),
                             scaling=True,
                             imputation=False,
                             imputation_nan_value=None,
                             feature_selection=False,
                             dim_reduction=True,
                             n_pca_components=10)
my_pipe.fit(X, y)



import numpy as np
import pandas as pd
from sklearn.datasets import make_classification
from sklearn.model_selection import StratifiedShuffleSplit

from photonai.modelwrapper.imbalanced_data_transformer import ImbalancedDataTransformer
from photonai.base import Hyperpipe, PipelineElement
from photonai.optimization import Categorical
from test.modelwrapper_tests.test_base_model_wrapper import BaseModelWrapperTest

from imblearn.over_sampling._smote.tests import test_smote
from imblearn.combine.tests import test_smote_tomek
from imblearn.under_sampling._prototype_selection.tests import test_instance_hardness_threshold


class ImbalancedDataTransformTest(BaseModelWrapperTest):
    """
    Tests based on implemented version by imblearn.
    For original implementation see:
    https://github.com/scikit-learn-contrib/imbalanced-learn/blob/master/imblearn/over_sampling/tests/
    https://github.com/scikit-learn-contrib/imbalanced-learn/tree/master/imblearn/over_sampling/_smote/tests
    """

    def setUp(self):
        super(ImbalancedDataTransformTest, self).setUp()
        self.model_wrapper = ImbalancedDataTransformer()

    def test_strategy(self):
        with self.assertRaises(ValueError):
            ImbalancedDataTransformer(method_name="something")

    def test_strategy_oversampling(self):
        """
        sample test of different functions based on imblearn implementation for oversampling methods.
        """
        imbalanced_data_transformer = ImbalancedDataTransformer(method_name='SMOTE',
                                                                config={"SMOTE": {"sampling_strategy": {0: 9, 1: 12},
                                                                                  "random_state": test_smote.RND_SEED}})

        # test_sample_regular_half() -> smote
        X_resampled, y_resampled = imbalanced_data_transformer.fit_transform(test_smote.X, test_smote.Y)
        X_gt = np.array([[0.11622591, -0.0317206], [0.77481731, 0.60935141],
                         [1.25192108, -0.22367336], [0.53366841, -0.30312976],
                         [1.52091956, -0.49283504], [-0.28162401, -2.10400981],
                         [0.83680821, 1.72827342], [0.3084254, 0.33299982],
                         [0.70472253, -0.73309052], [0.28893132, -0.38761769],
                         [1.15514042, 0.0129463], [0.88407872, 0.35454207],
                         [1.31301027, -0.92648734], [-1.11515198, -0.93689695],
                         [-0.18410027, -0.45194484], [0.9281014, 0.53085498],
                         [-0.14374509, 0.27370049], [-0.41635887, -0.38299653],
                         [0.08711622, 0.93259929], [1.70580611, -0.11219234], [0.36784496, -0.1953161]])
        y_gt = np.array([0, 1, 0, 0, 0, 1, 1, 1, 1, 1, 1, 0, 0, 1, 1, 1, 1, 0, 1, 0, 0])
        test_smote.assert_allclose(X_resampled, X_gt, rtol=test_smote.R_TOL)
        test_smote.assert_array_equal(y_resampled, y_gt)

    def test_strategy_undersampling(self):
        """
        sample test of different functions based on imblearn implementation for undersampling methods.
        """
        transformer = ImbalancedDataTransformer(method_name='InstanceHardnessThreshold',
                                                config={"InstanceHardnessThreshold":
                                                            {"estimator": test_instance_hardness_threshold.ESTIMATOR,
                                                             "sampling_strategy": {0: 6, 1: 8},
                                                             "random_state": test_instance_hardness_threshold.RND_SEED}
                                                        })

        X_resampled, y_resampled = transformer.fit_resample(test_instance_hardness_threshold.X,
                                                            test_instance_hardness_threshold.Y)
        assert X_resampled.shape == (15, 2)
        assert y_resampled.shape == (15,)

    def test_strategy_combine(self):
        """
        sample test of different functions based on imblearn implementation for oversampling methods.
        """
        imbalanced_data_transformer = ImbalancedDataTransformer(method_name='SMOTETomek',
                                                                config={"SMOTETomek":
                                                                            {"random_state": test_smote_tomek.RND_SEED}
                                                                        })
        X_resampled, y_resampled = imbalanced_data_transformer.fit_resample(test_smote_tomek.X, test_smote_tomek.Y)
        X_gt = np.array(
            [
                [0.68481731, 0.51935141],
                [1.34192108, -0.13367336],
                [0.62366841, -0.21312976],
                [1.61091956, -0.40283504],
                [-0.37162401, -2.19400981],
                [0.74680821, 1.63827342],
                [0.61472253, -0.82309052],
                [0.19893132, -0.47761769],
                [1.40301027, -0.83648734],
                [-1.20515198, -1.02689695],
                [-0.23374509, 0.18370049],
                [-0.00288378, 0.84259929],
                [1.79580611, -0.02219234],
                [0.38307743, -0.05670439],
                [0.70319159, -0.02571667],
                [0.75052536, -0.19246518],
            ]
        )
        y_gt = np.array([1, 0, 0, 0, 1, 1, 1, 1, 0, 1, 1, 1, 0, 0, 0, 0])
        test_smote_tomek.assert_allclose(X_resampled, X_gt, rtol=test_smote_tomek.R_TOL)
        test_smote_tomek.assert_array_equal(y_resampled, y_gt)

    def test_config_parameter(self):
        with self.assertRaises(ValueError):
             ImbalancedDataTransformer(method_name='SMOTETomek', config={"SMOTETomek": test_smote_tomek.RND_SEED})

    def test_different_strategies(self):
        def target_relative(y_true, y_pred):
            return (y_true == 0).sum() / len(y_true)

        def target_absolute(y_true, y_pred):
            return len(y_true)

        X, y = make_classification(weights=[0.9, 0.1], n_samples=300)

        my_pipe = Hyperpipe('balanced_pipe',
                            optimizer='grid_search',
                            metrics=['accuracy', target_relative, target_absolute],
                            best_config_metric="accuracy",
                            inner_cv=StratifiedShuffleSplit(n_splits=3, test_size=0.2),
                            ignore_sanity_checks=True,
                            verbosity=1,
                            project_folder='./tmp/')

        my_pipe += PipelineElement('StandardScaler')

        tested_methods = Categorical(['RandomOverSampler', 'RandomUnderSampler'])
        my_pipe += PipelineElement('ImbalancedDataTransformer',
                                   hyperparameters={'method_name': tested_methods},
                                   test_disabled=False)
        my_pipe += PipelineElement("LogisticRegression")
        my_pipe.fit(X, y)

        # -> test samples per strategy
        test_perf = pd.DataFrame([], columns=["config", "acc", "class_distribution", "absolute_samples"])

        for i, test_config_item in enumerate(my_pipe.results.outer_folds[0].tested_config_list):
            config = test_config_item.config_dict["ImbalancedDataTransformer__method_name"]
            acc = round(test_config_item.metrics_train[0].value, 3)
            relative = round(test_config_item.metrics_train[2].value, 3)
            absolute = round(test_config_item.metrics_train[4].value, 3)
            test_perf = pd.concat([test_perf, pd.DataFrame({'config': [config],
                                                            'acc': [acc],
                                                            'class_distribution': [relative],
                                                            'absolute_samples': [absolute]})], ignore_index=True)

        self.assertGreater(test_perf[test_perf["config"] == "RandomOverSampler"]["absolute_samples"].tolist()[0],
                           test_perf[test_perf["config"] == "RandomUnderSampler"]["absolute_samples"].tolist()[0])

        self.assertEqual(test_perf[test_perf["config"] == "RandomOverSampler"]["class_distribution"].tolist()[0],
                         test_perf[test_perf["config"] == "RandomUnderSampler"]["class_distribution"].tolist()[0])



from photonai import Hyperpipe
from sklearn.datasets import load_breast_cancer

X, y = load_breast_cancer(True)

# After optimization is finished, PHOTONAI saves the user's pipeline
# fitted with the best hyperparameter configuration found
# as "best_model.photonai" in the project's result folder.
# this is done automatically, however the use may do so manually by calling
# my_pipe.save_optimum_pipe('/home/photon_user/photon_test/optimum_pipe.photon')

my_pipe = Hyperpipe.load_optimum_pipe("full_path/to/best_model.photonai")
predictions = my_pipe.predict(X)

# get permutation importances posthoc
reloaded_hyperpipe = Hyperpipe.reload_hyperpipe("full_path/to/results_folder/", X, y)
post_hoc_perm_importances = Hyperpipe.get_permutation_feature_importances(n_repeats=5, random_state=0)



from photonai import RandomSearchOptimizer, IntegerRange, FloatRange, Categorical, PipelineElement, Hyperpipe
from sklearn.base import BaseEstimator, ClassifierMixin
from scipy.stats import pearsonr
import numpy as np

# ------------------------------------------------------------------------------------------------------
# Case 1: Using only random search optimizer to generate the config
# ------------------------------------------------------------------------------------------------------


class TVBDummyElement(BaseEstimator, ClassifierMixin):
    # todo: list tvb params with name here!
    def __init__(self, param1=None, param2=None, param3=None):
        # important: every hyperparameter needs to be stored in an instance variable
        self.param1 = param1
        self.param2 = param2
        self.param3 = param3

    def fit(self, X, y):
        return self

    def predict(self, X):
        return None


tvb_hyperparameters = {'param1': IntegerRange(0, 10),
                       'param2': FloatRange(0, 10, range_type="geomspace"),
                       'param3': Categorical(['uno', 'dos', 'tres'])
                       }

tvb_dummy = PipelineElement.create('tvb_dummy', TVBDummyElement(), hyperparameters=tvb_hyperparameters)

random_search_optimizer = RandomSearchOptimizer(n_configurations=25)
random_search_optimizer.prepare([tvb_dummy], maximize_metric=True)
for next_config in random_search_optimizer.ask:
    print(next_config)


# ------------------------------------------------------------------------------------------------------
# Case 2: Using photonai as Infrastructure
# will iterate each config the random search generator yields and:
# -> start the docker
# -> load the result
# -> compare the result to the ground truth for the specified metrics
# -> (log each config and all results)
# ------------------------------------------------------------------------------------------------------

class OnePersonNoCV:

    def init_(self):
        pass

    def split(self, X, y=None, groups=None):
        yield np.array([0]), np.array([0])


class TVBStarterElement(TVBDummyElement):

    def __init__(self, docker_container, param1=None, param2=None, param3=None):
        super(TVBStarterElement, self).__init__(param1, param2, param3)
        self.docker_container = docker_container
        self.params = [param1, param2, param3]
        self.outputfile = None

    def fit(self, X, y):
        self.return_value = y

        # # when the code reaches this place, param1, param2, param3 are already set by hyperparameter config.
        # cmd = 'Some TVB Command using param1 {}, param2 {}, param3 {}'.format(*self.params)
        # self.docker_container.exec_run(cmd)
        # # somehow monitor the process until command has finished....
        # self.outputfile = 'Docker Output File'
        # return self

    def predict(self, X):
        return self.return_value

        # outputdata = pd.read_csv(self.outputfile, delimiter=" ")
        # # convert to matrix
        # return outputdata


def score_tvb_output(X, y):
    # of course this works for one person only
    return pearsonr(np.squeeze(X), np.squeeze(y))[0]

# import docker
# https://docker-py.readthedocs.io/en/stable/index.html
# client = docker.from_env()
# tvb_container_id = 'fast-tbv-1'
# docker_container = client.containers.get(tvb_container_id)
docker_container = None

my_first_person_input_file = '/path/to/first/person'
my_first_person_target_file = np.array([0, 0, 1, 1, 0])

setup = Hyperpipe('tvb_optimizer_for_one_person',
                  outer_cv=None,
                  inner_cv=OnePersonNoCV(),  # when using more than one person use shuffle split
                  optimizer='random_search',
                  optimizer_params={'n_configurations': 25},
                  metrics=[('tvb_metric', score_tvb_output)],
                  best_config_metric='tvb_metric',
                  project_folder='./photonai_output_person_1',
                  allow_multidim_targets=True,
                  use_test_set=False,
                  verbosity=2)

setup += PipelineElement.create('tvb_item', TVBStarterElement(docker_container), hyperparameters=tvb_hyperparameters)
setup.fit([my_first_person_input_file], [my_first_person_target_file])







import logging
from photonai.base.hyperpipe import Hyperpipe, OutputSettings
from photonai.base.photon_elements import PipelineElement, Switch, Stack
from photonai.optimization.hyperparameters import IntegerRange
from photonai.optimization import FloatRange
from sklearn.model_selection import KFold, StratifiedKFold
from sklearn.base import BaseEstimator, ClassifierMixin, RegressorMixin
from sklearn.model_selection._split import BaseCrossValidator, BaseShuffleSplit, _RepeatedSplits
from photonai.photonlogger.logger import logger
from typing import Optional, List, Union
from pathlib import Path
import numpy as np
import pandas as pd


class ClassifierSwitch(Switch, ClassifierMixin):
    """The PHOTONAI ClassifierSwitch creates an instance of a PHOTONAI Switch with classification-specific estimators
    and pre-defined hyperparameter ranges.

        The ClassifierSwitch contains:
            - an SVC
                - C: [1e-8, 1e-6, 1e-4, 1e-2, 1, 1e2, 1e4, 1e6, 1e8]
                - kernel: ['rbf', 'linear'],
                - max_iter=1000
            - a RandomForestClassifier
                - max_features: ["sqrt", "log2"]
                - min_samples_leaf: [0.01, 0.1, 0.2]
            - an AdaBoostClassifier
                - n_estimators: [10, 25, 50]
            - LogisticRegression
                - C: [1e-4, 1e-2, 1, 1e2, 1e4]
                - penalty: ['l1', 'l2']
            - GaussianNB
            - KNeighborsClassifier
                - n_neighbors: [5, 10, 15]

              Attributes:

                  elements (list):
                      Contains `all PipelineElement or Hyperpipe objects that are added to the pipeline.
                      The RegressorSwitch contains an SVR, a RandomForestRegressor, LinearRegression, an AdaBoostRegressor
                      and KNeighborsRegressor.

              Example:
                  ``` python
                     from photonai import ClassificationPipe, ClassifierSwitch
                     from sklearn.datasets import load_breast_cancer

                     X, y = load_breast_cancer(return_X_y=True)
                     my_pipe = ClassificationPipe(name='breast_cancer_analysis',
                                                  add_estimator=False)
                     my_pipe += ClassifierSwitch()
                     # The default estimator is a ClassifierSwitch containing an SVC, a RandomForestClassifier,
                     # an AdaBoostClassifier, LogisticRegression, GaussianNB and KNeighborsClassifier.
                     my_pipe.fit(X, y)
                  ```
                  """

    def __init__(self, name: str = 'classifier_switch') -> None:
        element_list = [PipelineElement('SVC', hyperparameters={'C': [1e-8, 1e-6, 1e-4, 1e-2, 1, 1e2, 1e4, 1e6, 1e8],
                                                                'kernel': ['rbf', 'linear']}, max_iter=1000),
                        PipelineElement('RandomForestClassifier', hyperparameters={"max_features": ["sqrt", "log2"],
                                                                                   "min_samples_leaf": [0.01, 0.1, 0.2]}),
                        PipelineElement('AdaBoostClassifier', hyperparameters={'n_estimators': [10, 25, 50]}),
                        PipelineElement('LogisticRegression',
                                        hyperparameters={"C": [1e-4, 1e-2, 1, 1e2, 1e4],
                                                         "penalty": ['l1', 'l2']},
                                        solver='saga', n_jobs=1),
                        PipelineElement('GaussianNB'),
                        PipelineElement('KNeighborsClassifier', hyperparameters={"n_neighbors": [5, 10, 15]})]
        # todo: make a dedicated documentation site
        super(ClassifierSwitch, self).__init__(name, elements=element_list)


class RegressorSwitch(Switch, RegressorMixin):
    """The PHOTONAI RegressorSwitch creates an instance of a PHOTONAI Switch with regression-specific estimators and
    pre-defined hyperparameter ranges.

    The RegressorSwitch contains:
        - an SVR
            - C: [1e-8, 1e-6, 1e-4, 1e-2, 1, 1e2, 1e4, 1e6, 1e8]
            - kernel: ['rbf', 'linear'],
            - max_iter=1000
        - a RandomForestRegressor
            - max_features: ["sqrt", "log2"]
            - min_samples_leaf: [0.01, 0.1, 0.2]
        - an AdaBoostRegressor
            - n_estimators: [10, 25, 50]
        - LinearRegression
        - KNeighborsRegressor
            - n_neighbors: [5, 10, 15]


          Attributes:

              elements (list):
                  Contains `all PipelineElement or Hyperpipe objects that are added to the pipeline.
                  The RegressorSwitch contains an SVR, a RandomForestRegressor, LinearRegression, an AdaBoostRegressor
                  and KNeighborsRegressor.

          Example:
              ``` python
                from sklearn.datasets import load_diabetes
                from photonai import RegressionPipe, RegressorSwitch

                my_pipe = RegressionPipe('diabetes',
                                         add_estimator=False)
                # load data and train
                X, y = load_diabetes(return_X_y=True)
                my_pipe += RegressorSwitch()
                # The RegressorSwitch contains an SVR, a RandomForestRegressor, LinearRegression, an AdaBoostRegressor
                # and KNeighborsRegressor.
                my_pipe.fit(X, y)

              ```
              """

    def __init__(self, name: str = 'regressor_switch') -> None:
        element_list = [PipelineElement('SVR', hyperparameters={'C': [1e-8, 1e-6, 1e-4, 1e-2, 1, 1e2, 1e4, 1e6, 1e8],
                                                                'kernel': ['rbf', 'linear']}, max_iter=1000),
                        PipelineElement('RandomForestRegressor', hyperparameters={"max_features": ["sqrt", "log2"],
                                                                                  "min_samples_leaf": [0.01, 0.1,
                                                                                                       0.2]}),
                        PipelineElement('AdaBoostRegressor', hyperparameters={'n_estimators': [10, 25, 50]}),
                        PipelineElement('LinearRegression'),
                        PipelineElement('KNeighborsRegressor', hyperparameters={"n_neighbors": [5, 10, 15]})]

        super(RegressorSwitch, self).__init__(name, elements=element_list)


class DefaultPipeline(Hyperpipe):

    def __init__(self,
                 name: str = 'default_pipeline',
                 project_folder: Union[str, Path] = './',
                 inner_cv: Union[BaseCrossValidator, BaseShuffleSplit, _RepeatedSplits] = KFold(n_splits=10,
                                                                                                shuffle=True,
                                                                                                random_state=42),
                 outer_cv: Union[BaseCrossValidator, BaseShuffleSplit, _RepeatedSplits] = KFold(n_splits=5,
                                                                                                shuffle=True,
                                                                                                random_state=42),
                 metrics: list = None,
                 best_config_metric: str = '',
                 optimizer: str = 'random_grid_search',
                 optimizer_params: dict = {'n_configurations': 25},
                 X_csv_path: Union[str, Path, None] = None,
                 y_csv_path: Union[str, Path, None] = None,
                 delimiter=",",
                 add_default_pipeline_elements: bool = True,
                 scaling: bool = True,
                 imputation: bool = False,
                 imputation_nan_value=np.nan,
                 feature_selection: bool = False,
                 dim_reduction: bool = False,
                 n_pca_components: Union[int, float, None, IntegerRange, FloatRange] = None,
                 add_estimator: bool = True,
                 default_estimator: BaseEstimator = None,
                 **kwargs):

        self.X_csv_path = X_csv_path
        self.y_csv_path = y_csv_path
        self.delimiter = delimiter

        super(DefaultPipeline, self).__init__(name=name,
                                              project_folder=project_folder,
                                              inner_cv=inner_cv,
                                              outer_cv=outer_cv,
                                              metrics=metrics,
                                              best_config_metric=best_config_metric,
                                              optimizer=optimizer,
                                              optimizer_params=optimizer_params,
                                              **kwargs)

        if add_default_pipeline_elements is True:
            self.set_default_pipeline(scaling, imputation, imputation_nan_value, feature_selection, dim_reduction,
                                      n_pca_components, add_estimator, default_estimator)

    def set_default_pipeline(self, scaling, imputation, imputation_nan_value, feature_selection, dim_reduction,
                             n_pca_components, add_estimator, default_estimator):
        logger.stars()
        if scaling is True:
            logger.photon_system_log("USING STANDARD SCALER ")
            self.add(PipelineElement('StandardScaler'))
        if imputation is True:
            logger.photon_system_log("USING SIMPLE IMPUTER ")
            self.add(PipelineElement('SimpleImputer', missing_values=imputation_nan_value))
        if feature_selection:
            self.add(PipelineElement('FClassifSelectPercentile',
                                     hyperparameters={'percentile': IntegerRange(10, 50, step=10)},
                                     test_disabled=True))
        if dim_reduction is True:
            logger.photon_system_log("USING PCA ")
            self.add(PipelineElement('PCA', n_components=n_pca_components, test_disabled=True))
        if add_estimator is True:
            if default_estimator is None:
                raise ValueError("If paramater add_estimator is True, then default_estimator cannot be None. Use a "
                                 "string keyword argument such as 'SVC' or 'SVR' or photon native constructs "
                                 "such as PipelineElements, Switchs oder Stacks")
            ve = ValueError("Currently only string keywords such as 'SVC' or 'SVR' or photon native constructs "
                                 "such as PipelineElements, Switchs oder Stacks are allowed")
            if isinstance(default_estimator, str):
                estimator = PipelineElement(default_estimator)
            else:
                try:
                    estimator = default_estimator()
                    if not isinstance(estimator, (PipelineElement, Switch, Stack, ClassifierSwitch, RegressorSwitch)):
                        raise ve
                except Exception as e:
                    logging.error(e)
                    raise ve
            self += estimator
            logger.photon_system_log("USING " + self.elements[-1].name.upper())
            if hasattr(self.elements[-1], 'elements'):
                for idx, e in enumerate(self.elements[-1].elements):
                    logger.photon_system_log(e.name)
                    logger.photon_system_log(e.initial_hyperparameters)
                    logger.photon_system_log("---")
        logger.stars()

    def fit(self, X=None, y=None):
        if (X is not None and self.X_csv_path is not None) or (y is not None and self.y_csv_path is not None):
            raise ValueError("You can either give the fit function data or the pipe definition paths "
                             "to csv files to load data from. Not both.")

        if (X is None and self.X_csv_path is None) or (y is None and self.y_csv_path is None):
            raise ValueError("Please give either the fit function array-like X and y data "
                             "or give the pipe definition paths to csv files to load X and y data from.")

        X = X if X is not None else pd.read_csv(self.X_csv_path, delimiter=self.delimiter)
        y = y if y is not None else pd.read_csv(self.y_csv_path, delimiter=self.delimiter)
        super().fit(X, y)


class ClassificationPipe(DefaultPipeline):
    """The PHOTONAI ClassificationPipe creates an instance of a machine learning pipeline with classification-
    specific defaults, such as classification metrics. If not specified otherwise, a default pipeline is created
     consisting of a standard scaler, dimensionality reduction via PCA and an estimator switch, i.e. an OR-element
     comparing different learning algorithms for that pipeline.

     The user can prevent the creation of a default pipeline, and instead add a custom pipeline: So called PHOTONAI
     PipelineElements can be added, each of them being a data-processing method or a learning algorithm.
     By choosing and combining data-processing methods or algorithms, and arranging them with the PHOTONAI classes,
    simple and complex pipeline architectures can be designed rapidly.

    The PHOTONAI ClassificationPipe is a child of the Hyperpipe class and as such, automatizes the nested training,
    test and hyperparameter optimization procedures.

    The Hyperpipe monitors:
      - the nested-cross-validated training and test procedure,
      - communicates with the hyperparameter optimization strategy,
      - streams information between the pipeline elements,
      - logs all results obtained and evaluates the performance,
      - guides the hyperparameter optimization process by a so-called best config metric which is used to select
        the best performing hyperparameter configuration.

      Attributes:
          optimum_pipe (PhotonPipeline):
              An sklearn pipeline object that is fitted to the training data
              according to the best hyperparameter configuration found.
              Currently, we don't create an ensemble of all best hyperparameter
              configs over all folds. We find the best config by comparing
              the test error across outer folds. The hyperparameter config of the best
              fold is used as the optimal model and is then trained on the complete set.

          best_config (dict):
              Dictionary containing the hyperparameters of the
              best configuration. Contains the parameters in the sklearn
              interface of model_name__parameter_name: parameter value.

          results (MDBHyperpipe):
              Object containing all information about the for the
              performed hyperparameter search. Holds the training and test
              metrics for all outer folds, inner folds
              and configurations, as well as additional information.

          elements (list):
              Contains `all PipelineElement or Hyperpipe
              objects that are added to the pipeline.

      Example:
          ``` python
            from photonai import ClassificationPipe
            from sklearn.datasets import load_breast_cancer

            X, y = load_breast_cancer(return_X_y=True)
            my_pipe = ClassificationPipe(name='breast_cancer_analysis',
                                         add_default_pipeline_elements=True,
                                         scaling=True,
                                         imputation=False,
                                         imputation_nan_value=None,
                                         feature_selection=False,
                                         dim_reduction=True,
                                         n_pca_components=10,
                                         add_estimator=True)
            my_pipe.fit(X, y)
            # The default estimator is a ClassifierSwitch containing an SVC, a RandomForestClassifier,
            #                an AdaBoostClassifier, LogisticRegression, GaussianNB and KNeighborsClassifier.

          ```

      """

    def __init__(self,
                 name: Optional[str] = 'classification_pipeline',
                 project_folder: Union[str, Path] = './classification/',
                 metrics: list = None,
                 best_config_metric: str = 'balanced_accuracy',
                 default_estimator: Union[BaseEstimator, ClassifierSwitch] = ClassifierSwitch,
                 **kwargs):
        """
                Parameters:
                       name:
                           Name of RegressionPipe instance.

                       X_csv_path:
                           A path or string to a csv_file, which contains the feature values. If given, then the features
                           are loaded from that file in the fit function.
                           Default is None.

                       y_csv_path:
                           A path or string to a csv_file, which contains the target values. If given, then the target values
                           are loaded from that file in the fit function.
                           Default is None.

                       delimiter:
                           The delimiter that is used in both the csv files from X_csv_path and y_csv_path.
                           Default is ",".

                       add_default_pipeline_elements (bool):
                           If a default pipeline shall be generated. The default pipeline consists of
                            scaling with a StandardScaler, dimensionality reduction via PCA and an RegressionSwitch.
                            The default pipeline can be customized by the parameters
                            - scaling = True/False
                            - imputation = True/False
                            - imputation_nan_value = np.nan
                            - feature_selection = True/False
                            - dim_reduction = True/False
                            - n_pca_components = int, float, IntegerRange, FloatRange, None
                            - add_estimator = True/False
                            - default_estimator: BaseEstimator = RegressionSwitch

                        scaling:
                           If True, a StandardScaler is added to the default pipeline. Only relevant if parameter
                           add_default_pipelne is True.
                           Default is True

                        imputation:
                           If True, a SimpleImputer is added to the default pipeline. Only relevant if parameter
                               add_default_pipelne is True.
                           Default is False.

                        imputation_nan_value:
                           If imputation is set to True, this parameter defines the value that defines missing items.
                            Default is np.nan.

                        feature_selection:
                           If True, a feature selection element is added to the default pipeline. Only relevant if parameter
                               add_default_pipelne is True.
                           Default is False.

                        dim_reduction:
                           If True, a PCA is added to the default pipeline. Only relevant if parameter add_default_pipeline is
                           True.
                           Default is True.

                        n_pca_components:
                           If dim_reduction is True, this parameter defines the number of PCA components to be kept.
                           It can be an Integer, Float, an IntegerRange, FloatRange or None.
                           Default is None.

                        add_estimator:
                           If True, an estimator which is defined in default_estimator is added to the pipeline. Only relevant
                           if parameter add_default_pipelne is True.
                           Default is True.

                        default_estimator:
                           If parameter add_estimator is True, then an object of the type defined by default_estimator
                           is added to the pipeline. Use a string keyword argument such as 'SVC' or 'SVR'
                           or photon native constructs such as PipelineElement, Switch or Stack.

                           Default is ClassifierSwitch containing an SVC, a RandomForestClassifier,
                           an AdaBoostClassifier, LogisticRegression, GaussianNB and KNeighborsClassifier.

                       inner_cv:
                           Cross validation strategy to test hyperparameter configurations, generates the validation set.
                           Default is KFold(n_splits=10, shuffle=True, random_state=42).

                       outer_cv:
                           Cross validation strategy to use for the hyperparameter search itself, generates the test set.
                           Default is KFold(n_splits=5, shuffle=True, random_state=42).

                       optimizer:
                           Hyperparameter optimization algorithm.

                           - In case a string literal is given:
                               - "grid_search": Optimizer that iteratively tests all possible hyperparameter combinations.
                               - "random_grid_search": A variation of the grid search optimization that randomly picks
                                   hyperparameter combinations from all possible hyperparameter combinations.
                               - "sk_opt": Scikit-Optimize based on theories of bayesian optimization.
                               - "random_search": randomly chooses hyperparameter from grid-free domain.
                               - "smac": SMAC based on theories of bayesian optimization.
                               - "nevergrad": Nevergrad based on theories of evolutionary learning.

                           - In case an object is given:
                               expects the object to have the following methods:
                               - `ask`: returns a hyperparameter configuration in form of an dictionary containing
                                   key->value pairs in the sklearn parameter encoding `model_name__parameter_name: parameter_value`
                               - `prepare`: takes a list of pipeline elements and their particular hyperparameters to prepare the
                                            hyperparameter space
                               - `tell`: gets a tested config and the respective performance in order to
                                   calculate a smart next configuration to process

                           Default is random_search.

                       optimizer_params:
                           Define parameters of the optimizer. Given as dict.
                           Default is {'n_configurations': 25}

                       metrics:
                           Metrics that should be calculated for both training, validation and test set
                           Use the pre-imported metrics from sklearn and photonai, or register your own

                           - Metrics for `classification`:
                               - `accuracy`: sklearn.metrics.accuracy_score
                               - `matthews_corrcoef`: sklearn.metrics.matthews_corrcoef
                               - `confusion_matrix`: sklearn.metrics.confusion_matrix,
                               - `f1_score`: sklearn.metrics.f1_score
                               - `hamming_loss`: sklearn.metrics.hamming_loss
                               - `log_loss`: sklearn.metrics.log_loss
                               - `precision`: sklearn.metrics.precision_score
                               - `recall`: sklearn.metrics.recall_score
                           - Other metrics
                               - `pearson_correlation`: photon_core.framework.Metrics.pearson_correlation
                               - `variance_explained`:  photon_core.framework.Metrics.variance_explained_score
                               - `categorical_accuracy`: photon_core.framework.Metrics.categorical_accuracy_score

                           Default are ['balanced_accuracy', 'specificity', 'precision', 'recall', 'f1_score',
                                        'matthews_corrcoef'].

                       best_config_metric:
                           The metric that should be maximized or minimized in order to choose
                           the best hyperparameter configuration.

                           Default is 'balanced_accuracy'.

                       eval_final_performance:
                           DEPRECATED! Use "use_test_set" instead!

                       use_test_set:
                           If the metrics should be calculated for the test set,
                           otherwise the test set is seperated but not used.

                           Default is True.

                       project_folder:
                           The output folder in which all files generated by the
                           PHOTONAI project are saved to.

                       test_size:
                           The amount of the data that should be left out if no outer_cv is given and
                           eval_final_performance is set to True.

                           Default is 0.2.

                       calculate_metrics_per_fold:
                           If True, the metrics are calculated for each inner_fold.
                           If False, calculate_metrics_across_folds must be True.

                       calculate_metrics_across_folds:
                           If True, the metrics are calculated across all inner_fold.
                           If False, calculate_metrics_per_fold must be True.

                       ignore_sanity_checks:
                           If True, photonai will not verify use cases such as:
                               - classification, imbalanced classes and best_config_metric set to "accuracy"

                       random_seed:
                           Random Seed. This is also given to numpy and optimizer libraries.

                       verbosity:
                           The level of verbosity, 0 is least talkative and
                           gives only warn and error, 1 gives adds info and 2 adds debug.

                       learning_curves:
                           Enables learning curve procedure. Evaluate learning process over
                           different sizes of input. Depends on learning_curves_cut.

                       learning_curves_cut:
                           The tested relative cuts for data size.

                       performance_constraints:
                           Objects that indicate whether a configuration should
                           be tested further. For example, the inner fold of a config
                           does not perform better than the dummy performance.

                       permutation_id:
                           String identifier for permutation tests.

                       cache_folder:
                           Folder path for multi-processing.

                       nr_of_processes:
                           Determined the amount of simultaneous calculation of outer_folds.

                       multi_threading:
                           If true dask is used in multi threading mode, if false multi processing

                       allow_multidim_targets:
                           Allows multidimensional targets.

               """

        metrics = metrics if metrics is not None else ['balanced_accuracy',
                                                       'specificity',
                                                       'precision',
                                                       'recall',
                                                       'f1_score',
                                                       'matthews_corrcoef']

        super(ClassificationPipe, self).__init__(name=name,
                                                 project_folder=project_folder,
                                                 metrics=metrics,
                                                 best_config_metric=best_config_metric,
                                                 default_estimator=default_estimator,
                                                 **kwargs)


class RegressionPipe(DefaultPipeline):
    """The PHOTONAI RegressionPipe creates an instance of a machine learning pipeline with regression-specific defaults,
    such as the most common regression metrics. If not specified otherwise, a default pipeline is created
         consisting of a standard scaler, dimensionality reduction via PCA and an estimator switch, i.e. an OR-element
         comparing different learning algorithms for that pipeline.

         The user can prevent the creation of a default pipeline, and instead add a custom pipeline: So called PHOTONAI
         PipelineElements can be added, each of them being a data-processing method or a learning algorithm.
         By choosing and combining data-processing methods or algorithms, and arranging them with the PHOTONAI classes,
        simple and complex pipeline architectures can be designed rapidly.

        The PHOTONAI RegressionPipe is a child of the Hyperpipe class and as such, automatizes the nested training,
        test and hyperparameter optimization procedures.

        The Hyperpipe monitors:
          - the nested-cross-validated training and test procedure,
          - communicates with the hyperparameter optimization strategy,
          - streams information between the pipeline elements,
          - logs all results obtained and evaluates the performance,
          - guides the hyperparameter optimization process by a so-called best config metric which is used to select
            the best performing hyperparameter configuration.

          Attributes:
              optimum_pipe (PhotonPipeline):
                  An sklearn pipeline object that is fitted to the training data
                  according to the best hyperparameter configuration found.
                  Currently, we don't create an ensemble of all best hyperparameter
                  configs over all folds. We find the best config by comparing
                  the test error across outer folds. The hyperparameter config of the best
                  fold is used as the optimal model and is then trained on the complete set.

              best_config (dict):
                  Dictionary containing the hyperparameters of the
                  best configuration. Contains the parameters in the sklearn
                  interface of model_name__parameter_name: parameter value.

              results (MDBHyperpipe):
                  Object containing all information about the for the
                  performed hyperparameter search. Holds the training and test
                  metrics for all outer folds, inner folds
                  and configurations, as well as additional information.

              elements (list):
                  Contains `all PipelineElement or Hyperpipe
                  objects that are added to the pipeline.

          Example:
              ``` python
                from sklearn.datasets import load_diabetes
                from photonai import RegressionPipe

                my_pipe = RegressionPipe('diabetes',
                                         add_default_pipeline_elements=True,
                                         scaling=True,
                                         imputation=False,
                                         imputation_nan_value=None,
                                         feature_selection=False,
                                         dim_reduction=True,
                                         n_pca_components=10,
                                         add_estimator=True)
                # load data and train
                X, y = load_diabetes(return_X_y=True)
                my_pipe.fit(X, y)
                # Default is RegressorSwitch, containing an SVR, a RandomForestRegressor, LinearRegression,
                # an AdaBoostRegressor and KNeighborsRegressor.

              ```
          """

    def __init__(self,
                 name: Optional[str] = 'classification_pipeline',
                 project_folder: Union[str, Path] = './regression/',
                 metrics: list = None,
                 best_config_metric: str = 'mean_squared_error',
                 default_estimator: Union[BaseEstimator, RegressorSwitch] = RegressorSwitch,
                 **kwargs):
        """
         Parameters:
                name:
                    Name of RegressionPipe instance.

                X_csv_path:
                    A path or string to a csv_file, which contains the feature values. If given, then the features
                    are loaded from that file in the fit function.
                    Default is None.

                y_csv_path:
                    A path or string to a csv_file, which contains the target values. If given, then the target values
                    are loaded from that file in the fit function.
                    Default is None.

                delimiter:
                    The delimiter that is used in both the csv files from X_csv_path and y_csv_path.
                    Default is ",".

                add_default_pipeline_elements (bool):
                    If a default pipeline shall be generated. The default pipeline consists of
                     scaling with a StandardScaler, dimensionality reduction via PCA and an RegressionSwitch.
                     The default pipeline can be customized by the parameters
                     - scaling = True/False
                     - imputation = True/False
                     - imputation_nan_value = np.nan
                     - feature_selection = True/False
                     - dim_reduction = True/False
                     - n_pca_components = int, float, IntegerRange, FloatRange, None
                     - add_estimator = True/False
                     - default_estimator: BaseEstimator = RegressionSwitch

                 scaling:
                    If True, a StandardScaler is added to the default pipeline. Only relevant if parameter
                    add_default_pipelne is True.
                    Default is True

                 imputation:
                    If True, a SimpleImputer is added to the default pipeline. Only relevant if parameter
                        add_default_pipelne is True.
                    Default is False.

                 imputation_nan_value:
                    If imputation is set to True, this parameter defines the value that defines missing items.
                     Default is np.nan.

                 feature_selection:
                    If True, a feature selection element is added to the default pipeline. Only relevant if parameter
                        add_default_pipelne is True.
                    Default is False.

                 dim_reduction:
                    If True, a PCA is added to the default pipeline. Only relevant if parameter add_default_pipeline is
                    True.
                    Default is True.

                 n_pca_components:
                    If dim_reduction is True, this parameter defines the number of PCA components to be kept.
                    It can be an Integer, Float, an IntegerRange, FloatRange or None.
                    Default is None.

                 add_estimator:
                    If True, an estimator which is defined in default_estimator is added to the pipeline. Only relevant
                    if parameter add_default_pipelne is True.
                    Default is True.

                 default_estimator:
                    If parameter add_estimator is True, then an object of the type defined by default_estimator
                    is added to the pipeline. Use a string keyword argument such as 'SVC' or 'SVR'
                    or photon native constructs such as PipelineElement, Switch or Stack.

                    Default is RegressorSwitch, containing an SVR, a RandomForestRegressor, LinearRegression,
                    an AdaBoostRegressor and KNeighborsRegressor.

                inner_cv:
                    Cross validation strategy to test hyperparameter configurations, generates the validation set.
                    Default is KFold(n_splits=10, shuffle=True, random_state=42).

                outer_cv:
                    Cross validation strategy to use for the hyperparameter search itself, generates the test set.
                    Default is KFold(n_splits=5, shuffle=True, random_state=42).

                optimizer:
                    Hyperparameter optimization algorithm.

                    - In case a string literal is given:
                        - "grid_search": Optimizer that iteratively tests all possible hyperparameter combinations.
                        - "random_grid_search": A variation of the grid search optimization that randomly picks
                            hyperparameter combinations from all possible hyperparameter combinations.
                        - "sk_opt": Scikit-Optimize based on theories of bayesian optimization.
                        - "random_search": randomly chooses hyperparameter from grid-free domain.
                        - "smac": SMAC based on theories of bayesian optimization.
                        - "nevergrad": Nevergrad based on theories of evolutionary learning.

                    - In case an object is given:
                        expects the object to have the following methods:
                        - `ask`: returns a hyperparameter configuration in form of an dictionary containing
                            key->value pairs in the sklearn parameter encoding `model_name__parameter_name: parameter_value`
                        - `prepare`: takes a list of pipeline elements and their particular hyperparameters to prepare the
                                     hyperparameter space
                        - `tell`: gets a tested config and the respective performance in order to
                            calculate a smart next configuration to process

                    Default is random_search.

                optimizer_params:
                    Define parameters of the optimizer. Given as dict.
                    Default is {'n_configurations': 25}

                metrics:
                    Metrics that should be calculated for both training, validation and test set
                    Use the pre-imported metrics from sklearn and photonai, or register your own

                    - Metrics for `regression`:
                        - `mean_squared_error`: sklearn.metrics.mean_squared_error
                        - `mean_absolute_error`: sklearn.metrics.mean_absolute_error
                        - `explained_variance`: sklearn.metrics.explained_variance_score
                        - `r2`: sklearn.metrics.r2_score
                    - Other metrics
                        - `pearson_correlation`: photon_core.framework.Metrics.pearson_correlation
                        - `variance_explained`:  photon_core.framework.Metrics.variance_explained_score
                        - `categorical_accuracy`: photon_core.framework.Metrics.categorical_accuracy_score

                    Default are ['mean_absolute_error', 'mean_squared_error', 'explained_variance']

                best_config_metric:
                    The metric that should be maximized or minimized in order to choose
                    the best hyperparameter configuration.

                    Default is 'mean_squared_error'.

                eval_final_performance:
                    DEPRECATED! Use "use_test_set" instead!

                use_test_set:
                    If the metrics should be calculated for the test set,
                    otherwise the test set is seperated but not used.

                    Default is True.

                project_folder:
                    The output folder in which all files generated by the
                    PHOTONAI project are saved to.

                test_size:
                    The amount of the data that should be left out if no outer_cv is given and
                    eval_final_performance is set to True.

                    Default is 0.2.

                calculate_metrics_per_fold:
                    If True, the metrics are calculated for each inner_fold.
                    If False, calculate_metrics_across_folds must be True.

                calculate_metrics_across_folds:
                    If True, the metrics are calculated across all inner_fold.
                    If False, calculate_metrics_per_fold must be True.

                ignore_sanity_checks:
                    If True, photonai will not verify use cases such as:
                        - classification, imbalanced classes and best_config_metric set to "accuracy"

                random_seed:
                    Random Seed. This is also given to numpy and optimizer libraries.

                verbosity:
                    The level of verbosity, 0 is least talkative and
                    gives only warn and error, 1 gives adds info and 2 adds debug.

                learning_curves:
                    Enables learning curve procedure. Evaluate learning process over
                    different sizes of input. Depends on learning_curves_cut.

                learning_curves_cut:
                    The tested relative cuts for data size.

                performance_constraints:
                    Objects that indicate whether a configuration should
                    be tested further. For example, the inner fold of a config
                    does not perform better than the dummy performance.

                permutation_id:
                    String identifier for permutation tests.

                cache_folder:
                    Folder path for multi-processing.

                nr_of_processes:
                    Determined the amount of simultaneous calculation of outer_folds.

                multi_threading:
                    If true dask is used in multi threading mode, if false multi processing

                allow_multidim_targets:
                    Allows multidimensional targets.

        """

        metrics = metrics if metrics is not None else ['mean_absolute_error',
                                                       'mean_squared_error',
                                                       'explained_variance']

        super(RegressionPipe, self).__init__(name=name,
                                             project_folder=project_folder,
                                             metrics=metrics,
                                             best_config_metric=best_config_metric,
                                             default_estimator=default_estimator,
                                             **kwargs)



BEST_MODEL_FILE = 'best_model.photonai'
BEST_CONFIG_PREDICTIONS_FILE = 'best_config_predictions.csv'
RESULTS_FILE = 'photonai_results.json'
SUMMARY_FILE = 'photonai_summary.txt'
BLUEPRINT_FILE = '_optimum_pipe_blueprint.pkl'
META_FILE = '_optimum_pipe_meta.pkl'
HYPERPIPE_CONFIG_FILE = 'hyperpipe_config.json'


from sklearn.model_selection import KFold
from sklearn.datasets import load_breast_cancer
from photonai import Hyperpipe, PipelineElement

my_pipe = Hyperpipe('basic_svm_pipe',
                    inner_cv=KFold(n_splits=5),
                    outer_cv=KFold(n_splits=3),
                    metrics=['balanced_accuracy',
                             'precision',
                             'recall',
                             'accuracy'],
                    best_config_metric='balanced_accuracy',
                    project_folder='./tmp')

my_pipe += PipelineElement('StandardScaler')

my_pipe += PipelineElement('PCA',
                           n_components=10,
                           test_disabled=True)

my_pipe += PipelineElement('LogisticRegression')


X, y = load_breast_cancer(return_X_y=True)
my_pipe.fit(X, y)




from sklearn.datasets import load_diabetes
from sklearn.model_selection import KFold

from photonai import Hyperpipe, PipelineElement, MinimumPerformanceConstraint, \
    DummyPerformanceConstraint, BestPerformanceConstraint, IntegerRange

import matplotlib.pyplot as plt

my_pipe = Hyperpipe(name='constrained_forest_pipe',
                    optimizer='grid_search',
                    metrics=['mean_squared_error', 'mean_absolute_error', 'pearson_correlation'],
                    best_config_metric='mean_squared_error',
                    outer_cv=KFold(n_splits=3, shuffle=True),
                    inner_cv=KFold(n_splits=10),
                    use_test_set=True,
                    verbosity=1,
                    project_folder='./tmp',
                    # output_settings=OutputSettings(mongodb_connect_url="mongodb://localhost:27017/photon_results",
                    #                               save_output=True),
                    performance_constraints=[DummyPerformanceConstraint('mean_absolute_error'),
                                             MinimumPerformanceConstraint('pearson_correlation', 0.65, 'any'),
                                             BestPerformanceConstraint('mean_squared_error', 3, 'mean')])


my_pipe += PipelineElement('StandardScaler')
my_pipe += PipelineElement('RandomForestRegressor', hyperparameters={'n_estimators': IntegerRange(5, 50)})

X, y = load_diabetes(return_X_y=True)
my_pipe.fit(X, y)


## plot Scatter plot

train_df = my_pipe.results_handler.get_mean_train_predictions()
pred_df = my_pipe.results_handler.get_test_predictions()

max_value = int(max(max(pred_df['y_true']), max(pred_df['y_pred']), max(train_df['y_pred'])))

fig, main_axes = plt.subplots()
main_axes.plot(range(max_value), range(max_value), color='black')
test_set = main_axes.scatter(pred_df["y_true"], pred_df["y_pred"], label="Test")
train_set = main_axes.scatter(train_df["y_true"], train_df["y_pred"], label="Training")
main_axes.legend(handles=[test_set, train_set], loc='lower right')
main_axes.set_xlabel("y true")
main_axes.set_ylabel("y predicted")

plt.show()



pipeline = initalize_pipeline()
data = sanity_check_input_data()

for outer_fold in outer_folds:
    outer_fold_data = outer_cv_strategy.get(outer_fold, data)

    # apply most trivial prediction strategy to estimate baseline performance
    run_dummy_estimator(outer_fold_data)

    # initialize hyperparameter optimization space
    hyperparameter_optimizer.prepare(pipeline)

    # ask hyperparameter optimization strategy
    # for next hyperparameter configuration
    for hp_config in hyperparameter_optimizer.ask():
        for inner_fold in inner_folds:
            inner_fold_data = inner_cv_strategy.get(inner_fold,
                                                    outer_fold_data)
            # train and evaluate on validation set
            current_performance = train_and_test_pipeline(hp_config,
                                                          pipeline,
                                                          inner_fold_data)

            # inform hyperparameter optimization strategy
            hyperparameter_optimizer.tell(current_performance)

            if performance_constraints:
                # check if hp_config shall be further evaluated
                # or is dismissed due to bad performance
                if not current_performance > performance_constraints:
                    break

        # log best hyperparameter configuration so far
        if current_performance > best_performance:
            best_performance = current_performance
            best_config = hp_config

    best_configs_outer_folds.append(best_config)

    # evaluate performance of best configuration on test set
    train_and_test_pipeline(best_config,
                            pipeline,
                            outer_fold_data)

# select overall best config across best configs of outer folds
overall_best_config = max_performance(best_configs_outer_folds)
# setup pipeline with best config
pipeline.set_params(overall_best_config)
# train with all data
pipeline.metrics_train(data)
# save the final model in a standardized format
pipeline.save()



from photonai.base import ClassificationPipe, ClassifierSwitch, RegressionPipe, RegressorSwitch
from photonai.helper.photon_base_test import elements_to_dict, PhotonBaseTest
from sklearn.model_selection import ShuffleSplit
from sklearn.svm import SVC
from sklearn.datasets import load_breast_cancer
import os
from pathlib import Path


class ModelZooTests(PhotonBaseTest):

    @classmethod
    def setUpClass(cls) -> None:
        cls.file = __file__
        super(ModelZooTests, cls).setUpClass()

    def test_kwargs_setup(self):
        name = 'super_duper_classifier'
        cp = ClassificationPipe(name=name,
                                inner_cv=ShuffleSplit(n_splits=3),
                                use_test_set=False,
                                dim_reduction=True)

        # Hyperpipe should have gotten inner_cv and use_test_set,
        # DefaultPipeline should have gotten dim_reduction
        self.assertEqual(cp.cross_validation.inner_cv.n_splits, ShuffleSplit(n_splits=3).n_splits)
        self.assertFalse(cp.cross_validation.use_test_set)
        self.assertEqual(cp.elements[1].name, 'PCA')

    def test_default_pipeline_setup(self):
        cp1 = ClassificationPipe(add_default_pipeline_elements=True,
                                 scaling=True,
                                 dim_reduction=True,
                                 n_pca_components=10,
                                 feature_selection=True,
                                 imputation=True,
                                 imputation_nan_value=-99,
                                 add_estimator=True,
                                 default_estimator='SVC')
        self.assertEqual(cp1.elements[0].name, 'StandardScaler')
        self.assertEqual(cp1.elements[1].name, 'SimpleImputer')
        self.assertEqual(cp1.elements[2].name, 'FClassifSelectPercentile')
        self.assertEqual(cp1.elements[3].name, 'PCA')
        self.assertEqual(cp1.elements[3].base_element.n_components, 10)
        self.assertEqual(cp1.elements[4].name, 'SVC')

        cp2 = ClassificationPipe(add_default_pipeline_elements=False,
                                 scaling=True,
                                 dim_reduction=True,
                                 n_pca_components=10,
                                 feature_selection=True,
                                 imputation=True,
                                 imputation_nan_value=-99,
                                 add_estimator=True,
                                 default_estimator='SVC')

        self.assertEqual(len(cp2.elements), 0)

        with self.assertRaises(ValueError):
            cp_svc = ClassificationPipe(default_estimator=SVC)

    def test_data_loading(self):
        test_directory = os.path.join(os.path.dirname(os.path.abspath(__file__)).split('test')[0], 'test')
        data_folder = Path(test_directory, 'base_tests/custom_elements/')
        rp_path = RegressionPipe(name='my_custom_regression_pipe',
                                 X_csv_path=data_folder.joinpath('breast_cancer_X.csv'),
                                 y_csv_path=data_folder.joinpath('breast_cancer_y.csv'))
        rp_path.fit()

        X, y = load_breast_cancer(return_X_y=True)
        with self.assertRaises(ValueError):
            rp_path.fit(X, y)

        with self.assertRaises(ValueError):
            rp_none = RegressionPipe()
            rp_none.fit()

        rp_numpy = RegressionPipe(name='custom_pipe2')
        rp_numpy.fit(X, y)

    def test_classifier_defaults(self):
        cl = ClassificationPipe()
        for m in ['balanced_accuracy', 'precision', 'recall']:
            self.assertTrue(m in cl.optimization.metrics)
        self.assertEqual(cl.optimization.best_config_metric, 'balanced_accuracy')
        self.assertIsInstance(cl.elements[-1], ClassifierSwitch)

    def test_regressor_defaults(self):
        reg = RegressionPipe()
        for r in ['mean_squared_error', 'mean_absolute_error', 'explained_variance']:
            self.assertTrue(r in reg.optimization.metrics)
        self.assertEqual(reg.optimization.best_config_metric, 'mean_squared_error')
        self.assertIsInstance(reg.elements[-1], RegressorSwitch)




# import unittest
# import numpy as np
# import warnings
# import glob
# from shutil import rmtree
#
# from sklearn.metrics import accuracy_score
#
# from photonai.helper.helper import PhotonDataHelper
# from photonai.base import PipelineElement, Switch, Hyperpipe
# from photonai.base.photon_pipeline import PhotonPipeline
# from sklearn.datasets import fetch_olivetti_faces
# from sklearn.model_selection import ShuffleSplit
# from photonai.optimization import FloatRange, Categorical, IntegerRange
# from photonai.optimization.hyperparameters import NumberRange
# from photonai.optimization.smac.smac import SMACOptimizer
# import photonai.optimization.smac.smac as photonai_smac
#
# if photonai_smac.__found__:
#     # Import ConfigSpace and different types of parameters
#     from smac.configspace import ConfigurationSpace
#     from ConfigSpace.hyperparameters import CategoricalHyperparameter, \
#         UniformFloatHyperparameter, UniformIntegerHyperparameter
#     from ConfigSpace.conditions import InCondition
#     # Import SMAC-utilities
#     from smac.scenario.scenario import Scenario
#     from smac.facade.smac_bb_facade import SMAC4BB as SMAC4BO
#     from smac.facade.smac_hpo_facade import SMAC4HPO
#     from smac.facade.smac_ac_facade import SMAC4AC
#     from smac.facade.smac_mf_facade import SMAC4MF as BOHB4HPO
#
#
# @unittest.skipIf(not photonai_smac.__found__, 'smac not available')
# class Smac3IntegrationTest(unittest.TestCase):
#
#     def setUp(self):
#         self.s_split = ShuffleSplit(n_splits=3, test_size=0.2, random_state=42)
#
#         self.time_limit = 20
#
#         self.pipe = Hyperpipe('basic_svm_pipe',
#                               optimizer='smac',
#                               optimizer_params={'facade': SMAC4BO,
#                                                 'wallclock_limit': self.time_limit,
#                                                 'rng': 42},
#                               metrics=['accuracy'],
#                               random_seed=True,
#                               best_config_metric='accuracy',
#                               inner_cv=self.s_split,
#                               verbosity=0,
#                               project_folder='./tmp/')
#
#     def simple_classification(self):
#         dataset = fetch_olivetti_faces(download_if_missing=True)
#         self.X = dataset["data"]
#         self.y = dataset["target"]
#         return self.X, self.y
#
#     # integration test for simple pipeline without Switch
#     def test_photon_implementation_simple(self):
#         # PHOTON implementation
#         # todo: debug !
#         # self.pipe.add(PipelineElement('StandardScaler'))
#         # self.pipe += PipelineElement('PCA', hyperparameters={'n_components': IntegerRange(5, 30)})
#         # self.pipe += PipelineElement('SVC', hyperparameters={'kernel': Categorical(["rbf", 'poly']),
#         #                                                      'C': FloatRange(0.5, 200)}, gamma='auto')
#         # self.X, self.y = self.simple_classification()
#         # self.pipe.fit(self.X, self.y)
#         #
#         # # direct AUTO ML implementation
#         # # Build Configuration Space which defines all parameters and their ranges
#         # cs = ConfigurationSpace()
#         # n_components = UniformIntegerHyperparameter("PCA__n_components", 5, 30)
#         # cs.add_hyperparameter(n_components)
#         # kernel = CategoricalHyperparameter("SVC__kernel", ["rbf", 'poly'])
#         # cs.add_hyperparameter(kernel)
#         # c = UniformFloatHyperparameter("SVC__C", 0.5, 200)
#         # cs.add_hyperparameter(c)
#         #
#         # # Scenario object
#         # scenario = Scenario({"run_obj": "quality",
#         #                      "cs": cs,
#         #                      "deterministic": "true",
#         #                      "wallclock_limit": self.time_limit,
#         #                      "limit_resources": False,
#         #                      'abort_on_first_run_crash': False
#         #                      })
#         #
#         # # Optimize, using a SMAC directly
#         # smac = SMAC4BO(scenario=scenario, rng=42, tae_runner=self.objective_function_simple)
#         # _ = smac.optimize()
#         #
#         # runhistory_photon = [x.metrics_test[0].value for x in self.pipe.results.outer_folds[0].tested_config_list]
#         # runhistory_original = [1-x for x in list(smac.solver.runhistory._cost_per_config.values())]
#         #
#         # min_len = min(len(runhistory_original), len(runhistory_photon))
#         # np.testing.assert_almost_equal(runhistory_photon[:min_len], runhistory_original[:min_len], 1)
#
#     def objective_function_simple(self, cfg):
#         cfg = {k: cfg[k] for k in cfg if cfg[k]}
#         values = []
#
#         train_indices = list(self.pipe.cross_validation.outer_folds.values())[0].train_indices
#         self._validation_X, self._validation_y, _ = PhotonDataHelper.split_data(self.X, self.y,
#                                                                                 kwargs=None, indices=train_indices)
#
#         for inner_fold in list(list(self.pipe.cross_validation.inner_folds.values())[0].values()):
#             sc = PipelineElement("StandardScaler", {})
#             pca = PipelineElement("PCA", {}, random_state=42)
#             svc = PipelineElement("SVC", {}, random_state=42, gamma='auto')
#             my_pipe = PhotonPipeline([('StandardScaler', sc), ('PCA', pca), ('SVC', svc)])
#             my_pipe.set_params(**cfg)
#             my_pipe.fit(self._validation_X[inner_fold.train_indices, :],
#                         self._validation_y[inner_fold.train_indices])
#             values.append(accuracy_score(self._validation_y[inner_fold.test_indices],
#                                          my_pipe.predict(self._validation_X[inner_fold.test_indices, :])
#                                          )
#                           )
#         return 1 - np.mean(values)
#
#     def test_further_parameters(self):
#         # todo: debug !
#         # n_configurations = 4
#         # pipe = Hyperpipe('basic_svm_pipe',
#         #                  optimizer='smac',
#         #                  optimizer_params={'facade': SMAC4BO,
#         #                                    'wallclock_limit': 1000,
#         #                                    'ta_run_limit': n_configurations,
#         #                                    'rng': 42},
#         #                  metrics=['accuracy'],
#         #                  random_seed=True,
#         #                  best_config_metric='accuracy',
#         #                  inner_cv=self.s_split,
#         #                  verbosity=0,
#         #                  project_folder='./tmp/')
#         #
#         # pipe.add(PipelineElement('StandardScaler'))
#         # pipe += PipelineElement('PCA', hyperparameters={'n_components': IntegerRange(5, 30)})
#         # pipe += PipelineElement('SVC', hyperparameters={'kernel': Categorical(["rbf", 'poly']),
#         #                                                      'C': FloatRange(0.5, 200)}, gamma='auto')
#         # X, y = self.simple_classification()
#         # pipe.fit(X, y)
#         # self.assertEqual(len(pipe.results.outer_folds[0].tested_config_list), n_configurations)
#
#     # integration test for pipeline with Switch
#     def test_photon_implementation_switch(self):
#         # todo: debug !
#         pass
#         # # PHOTON implementation
#         # self.pipe.add(PipelineElement('StandardScaler'))
#         # self.pipe += PipelineElement('PCA', hyperparameters={'n_components': IntegerRange(5, 30)})
#         # estimator_siwtch = Switch("Estimator")
#         # estimator_siwtch += PipelineElement('SVC', hyperparameters={'kernel': Categorical(["rbf", 'poly']),
#         #                                                             'C': FloatRange(0.5, 200)}, gamma='auto')
#         # estimator_siwtch += PipelineElement('RandomForestClassifier',
#         #                                     hyperparameters={'criterion': Categorical(['gini', 'entropy']),
#         #                                                      'min_samples_split': IntegerRange(2, 4)
#         #                                                      })
#         # self.pipe += estimator_siwtch
#         # self.X, self.y = self.simple_classification()
#         # self.pipe.fit(self.X, self.y)
#         #
#         # # direct AUTO ML implementation
#         #
#         # # Build Configuration Space which defines all parameters and their ranges
#         # cs = ConfigurationSpace()
#         # n_components = UniformIntegerHyperparameter("PCA__n_components", 5, 30)
#         # cs.add_hyperparameter(n_components)
#         #
#         # switch = CategoricalHyperparameter("Estimator_switch", ['svc', 'rf'])
#         # cs.add_hyperparameter(switch)
#         #
#         # kernel = CategoricalHyperparameter("SVC__kernel", ["rbf", 'poly'])
#         # cs.add_hyperparameter(kernel)
#         # c = UniformFloatHyperparameter("SVC__C", 0.5, 200)
#         # cs.add_hyperparameter(c)
#         # use_svc_c = InCondition(child=kernel, parent=switch, values=["svc"])
#         # use_svc_kernel = InCondition(child=c, parent=switch, values=["svc"])
#         #
#         # criterion = CategoricalHyperparameter("RandomForestClassifier__criterion", ['gini', 'entropy'])
#         # cs.add_hyperparameter(criterion)
#         # minsplit = UniformIntegerHyperparameter("RandomForestClassifier__min_samples_split", 2, 4)
#         # cs.add_hyperparameter(minsplit)
#         #
#         # use_rf_crit = InCondition(child=criterion, parent=switch, values=["rf"])
#         # use_rf_minsplit = InCondition(child=minsplit, parent=switch, values=["rf"])
#         #
#         # cs.add_conditions([use_svc_c, use_svc_kernel, use_rf_crit, use_rf_minsplit])
#         #
#         # # Scenario object
#         # scenario = Scenario({"run_obj": "quality",
#         #                      "cs": cs,
#         #                      "deterministic": "true",
#         #                      "wallclock_limit": self.time_limit,
#         #                      "limit_resources": False,
#         #                      'abort_on_first_run_crash': False
#         #                      })
#         #
#         # # Optimize, using a SMAC directly
#         # smac = SMAC4BO(scenario=scenario, rng=42, tae_runner=self.objective_function_switch)
#         # _ = smac.optimize()
#         #
#         # runhistory_photon = [x.metrics_test[0].value for x in self.pipe.results.outer_folds[0].tested_config_list]
#         # runhistory_original = [1 - x for x in list(smac.solver.runhistory._cost_per_config.values())]
#         #
#         # min_len = min(len(runhistory_original), len(runhistory_photon))
#         # np.testing.assert_allclose(runhistory_photon[:min_len], runhistory_original[:min_len], rtol=0.50)
#
#     def objective_function_switch(self, cfg):
#         cfg = {k: cfg[k] for k in cfg if cfg[k]}
#         values = []
#
#         train_indices = list(self.pipe.cross_validation.outer_folds.values())[0].train_indices
#         self._validation_X, self._validation_y, _ = PhotonDataHelper.split_data(self.X, self.y,
#                                                                                 kwargs=None,
#                                                                                 indices=train_indices)
#
#         switch = cfg["Estimator_switch"]
#         del cfg["Estimator_switch"]
#         for inner_fold in list(list(self.pipe.cross_validation.inner_folds.values())[0].values()):
#             sc = PipelineElement("StandardScaler", {})
#             pca = PipelineElement("PCA", {}, random_state=42)
#             if switch == 'svc':
#                 est = PipelineElement("SVC", {}, random_state=42, gamma='auto')
#                 name = 'SVC'
#             else:
#                 est = PipelineElement("RandomForestClassifier", {}, random_state=42)
#                 name = "RandomForestClassifier"
#             my_pipe = PhotonPipeline([('StandardScaler', sc), ('PCA', pca), (name, est)])
#             my_pipe.set_params(**cfg)
#             my_pipe.fit(self._validation_X[inner_fold.train_indices, :],
#                         self._validation_y[inner_fold.train_indices])
#             values.append(accuracy_score(self._validation_y[inner_fold.test_indices],
#                                          my_pipe.predict(self._validation_X[inner_fold.test_indices, :])
#                                          )
#                           )
#         return 1 - np.mean(values)
#
#     def test_facade(self):
#         config_space = ConfigurationSpace()
#         n_components = UniformIntegerHyperparameter("PCA__n_components", 5, 30)
#         config_space.add_hyperparameter(n_components)
#         scenario_dict = {"run_obj": "quality",
#                          "deterministic": "true",
#                          "cs": config_space,
#                          "wallclock_limit": 60
#                          }
#
#         with self.assertRaises(ValueError):
#             SMACOptimizer(facade="SMAC4BOO", scenario_dict=scenario_dict)
#
#         with self.assertRaises(ValueError):
#             facade = SMAC4BO(scenario=Scenario(scenario_dict))
#             SMACOptimizer(facade=facade, scenario_dict=scenario_dict)
#
#         facades = ["SMAC4BO", SMAC4BO, "SMAC4AC", SMAC4AC, "SMAC4HPO", SMAC4HPO, "BOHB4HPO", BOHB4HPO]
#         for facade in facades:
#             SMACOptimizer(facade=facade, scenario_dict=scenario_dict)
#
#     def test_other(self):
#         opt = SMACOptimizer(facade="SMAC4BO", intensifier_kwargs={'min_chall': 2})
#         self.assertIsNotNone(opt.intensifier_kwargs)
#
#         pipeline_elements = [PipelineElement('SVC', hyperparameters={'kernel': Categorical(["rbf", 'poly', "sigmoid"]),
#                                                                      'C': [0.6]})]
#
#         def of(x):
#             return x ** 2
#         with warnings.catch_warnings(record=True) as w:
#             opt.prepare(pipeline_elements=pipeline_elements, maximize_metric=True, objective_function=of)
#             assert any("PHOTONAI has detected some" in s for s in [e.message.args[0] for e in w])
#
#         pipeline_elements = [PipelineElement("SVC", hyperparameters={'C': FloatRange(0.1, 0.5, range_type='geomspace'),
#                                                                      'kernel': ['rbf']})]
#         opt = SMACOptimizer(facade="SMAC4BO")
#         with self.assertRaises(NotImplementedError):
#             opt.prepare(pipeline_elements=pipeline_elements, maximize_metric=True, objective_function=of)
#
#         pipeline_elements = [PipelineElement("SVC", hyperparameters={'C': NumberRange(1, 3, range_type='range')})]
#         opt = SMACOptimizer(facade="SMAC4BO")
#         with self.assertRaises(ValueError):
#             opt.prepare(pipeline_elements=pipeline_elements, maximize_metric=True, objective_function=of)
#
#     @classmethod
#     def tearDownClass(cls) -> None:
#         dirs = glob.glob("./smac3-output*/")
#         for dir_name in dirs:
#             rmtree(dir_name, ignore_errors=True)
